from pymoo.problems.single.ackley import *
from pymoo.problems.single.cantilevered_beam import *
from pymoo.problems.single.griewank import *
from pymoo.problems.single.knapsack import *
from pymoo.problems.single.pressure_vessel import *
from pymoo.problems.single.rastrigin import *
from pymoo.problems.single.rosenbrock import *
from pymoo.problems.single.schwefel import *
from pymoo.problems.single.sphere import *
from pymoo.problems.single.zakharov import *
from pymoo.problems.single.g import *
from pymoo.problems.single.himmelblau import *